/**
 * @file rpointer.h
 * @brief defines the relative Pointers 
 * @author <A href="http://www.infnet.verein.de/~rw/">Rudolf Weber</A> 2000 
 * nach einer Idee 1995
 *
 * Anderungen:
 *  16.3.2002: template RPointer bekommt zweiten Parameter SIZE, mit default
 *  27.3.2002: der off anteil von RPointer muss eine Referenz sein, dann 
 *             kann man gleich einen Persistent-Zeiger verwalten.
 *   Idee: ein relativer persitenter Pointer ist im wesentlichen ein Offset.
 *         Zur schoeneren Verwaltung gibt es eine Zugriffsstruktur, deren 
 *         Zuweisung die Segmentbasis abzieht.
 *  30.3.2002: off von Referenz zum Pointer, damit man auch leer initialisieren
 *             kann.
 *  29.11.2003: Operatoren -> und * 
 */
#ifndef RPOINTER_H
#define RPOINTER_H
#include<assert.h>
#include"mseg.h"
#include<sys/types.h>

namespace RAMS
{

 /**
  * \brief Generischer Relativer Zeiger
  *
  * Ein Relativer Zeiger ist immer abhaengig von einem Speichersegment
  */
 class RPointer_s
 {
 protected:
     /// Zugeh&ouml;riges Speichersegment
    Mseg *seg;
     /// Offset in diesem Speichersegment
    off_t* off;
    /// Speicher fuer offset
    off_t  soff; 
 public:
    RPointer_s() { seg=(Mseg *)NULL;off=(off_t *)NULL; }
    /**
     * Initialisierung
     */
    RPointer_s(Mseg *s,off_t& o):seg(s),off(&o) {}
    RPointer_s(Mseg *s,off_t& o,off_t koff):seg(s),off(&o) { *off=koff; }
    /// Initialisierung mit konstanten Offset
    RPointer_s(off_t koff,Mseg *s):seg(s),off(&soff),soff(koff) {}
    void init(Mseg *s,off_t& o) { seg=s; off=&o; }
    void init(off_t koff,Mseg *s) { off=&soff; soff=koff; seg=s; }
    off_t offset() const { return *off; }
    
    /**
     * Pr&uuml;fung, ob Speicher in das Segment pa&szlig;t
     */
    int ok() { return *off>=0 && *off < off_t(seg->length()); }
    /**
     * Gr&ouml;ssenberechnung
     */
    size_t operator-(RPointer_s a)
	{
	    assert(a.seg->begin()==seg->begin());
	    return (*off-*(a.off));
	}

    /// for special purposes
    void setoff(off_t o) { *off=o; }
    /**
     * Zuweisung eines relativen Zeiger in einem anderen Typ
     */
    RPointer_s& cast(const RPointer_s &anderer)
    {
	seg=anderer.seg;
	*off=*(anderer.off);
	return *this;
    }
    
    /**
     * Zuweisung eines absoluten Zeiger an unsern relativen Zeiger
     */
    void *operator=(void *p)
    {
      *off=(char *)p - (char *)(seg->begin());
      assert(ok());
      return p;
    }
    
    /**
     * Vergleichsoperator
     */
    bool operator==(const RPointer_s& other) const
    {
      return (seg->begin()==other.seg->begin() && offset() == other.offset());
    }
 };

 /**
  * @brief Relativer Zeiger auf einen Typ T
  */
 template<class T,int SIZE=sizeof(T)> class RPointer : public RPointer_s
 {
 public:
    RPointer():RPointer_s() {}
    RPointer(Mseg *s,off_t& o):RPointer_s(s,o) {}
    RPointer(Mseg *s,off_t& o,off_t koff):RPointer_s(s,o,koff) {}
    RPointer(off_t o,Mseg *s):RPointer_s(o,s) {}
    RPointer(const RPointer_s& a) { cast(a); }
    RPointer(Mseg *s,T* ap):RPointer_s((char *)ap - (char *)(s->begin()),s) {}
    RPointer(Mseg *s,off_t &o,T* ap)
     :RPointer(s,o,(char *)ap - (char *)(s->begin())) {}
    T *ptr() { return (T *)((char *)(seg->begin()) + *off); }
    T *operator->() { return ptr(); }
    T& operator*()  { return *ptr(); }
    RPointer<T,SIZE>& operator=(RPointer<T,SIZE>& anderer)
    {
      if(seg == 0)
      { // Initialisierung 
	seg=anderer.seg;
	off=&soff;
	*off=*(anderer.off);
      }
      else if(seg->begin()==anderer.seg->begin())
      {
	*off=*(anderer.off);
      }
      else
      {
	// wie anderer Zeiger
	*off=(char *)(anderer.ptr()) - (char *)(seg->begin());
      }
      assert(ok());
      return *this;
    }
    T *operator+=(off_t dazu)
    { *off+=dazu*SIZE;
      assert(ok());
      return ptr();
    }
    T *operator-=(off_t dazu)
    { *off-=dazu*SIZE;
      assert(*off >= 0);
      return ptr();
    }
    /**
     * Zuweisung eines absoluten Zeiger an unsern relativen Zeiger
     */
    T *operator=(T *p)
    {
      *off=(char *)p - (char *)(seg->begin());
      assert(ok());
      return p;
    }
 };

 typedef RPointer<char> RcharPointer;
};

/**
 * @brief Alignment auf 2er Potenzen
 *
 * @param  a 2er Potenz 8,16,...,4096,...
 * @param  w Wert, der aligned werden soll
 */
inline long align2(long a,long w)
{
   long e= (w & ~(a-1));
   if(e == w)  return e;
   else        return e+a;
}
#endif





